from typing import Any


class HarambeException(Exception):
    """Base exception for all custom exceptions in Harambe."""

    pass


class UnknownHTMLConverter(HarambeException):
    def __init__(self, converter_type: Any) -> None:
        super().__init__(f"Unknown HTML converter type: {converter_type}")


class GotoError(HarambeException):
    def __init__(self, url: str, status: int) -> None:
        super().__init__(
            f'Error calling goto("{url}"). Received unexpected status code: {status}'
        )


class SchemaValidationError(HarambeException):
    def __init__(self, message: str = None):
        super().__init__(message)


class CaptchaError(HarambeException):
    def __init__(self, message: str = "CAPTCHA was hit.") -> None:
        super().__init__(message)



class Schemas:
    regulation_documents = {
        "title": {
            "type": "string",
        },
        "document_url": {
            "type": "string",
        },
    }
    government_contracts = {
        "id": {
            "type": "string",
        },
        "title": {
            "type": "string",
        },
        "status": {
            "type": "string",
        },
        "description": {
            "type": "string",
        },
        "location": {
            "type": "string",
        },
        "type": {
            "type": "string",
        },
        "category": {
            "type": "string",
        },
        "posted_date": {
            "type": "string",
        },
        "due_date": {
            "type": "string",
        },
        "buyer_name": {
            "type": "string",
        },
        "buyer_contact_name": {
            "type": "string",
        },
        "buyer_contact_email": {
            "type": "string",
        },
        "buyer_contact_number": {
            "type": "string",
        },
        "attachments": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"title": {"type": "string"}, "url": {"type": "string"}},
            },
        },
        "procurement_items": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "code_type": {"type": "string"},
                    "code": {"type": "string"},
                    "code_description": {"type": "string"},
                    "description": {"type": "string"},
                },
            },
        },
    }

    government_contracts_small = {
        "id": {
            "type": "string",
            "required": True,
        },
        "title": {
            "type": "string",
            "required": True,
        },
        "description": {
            "type": "string",
        },
        "location": {
            "type": "string",
        },
        "type": {
            "type": "string",
        },
        "category": {
            "type": "string",
        },
        "posted_date": {
            "type": "string",
        },
        "due_date": {
            "type": "string",
        },
        "agency": {
            "type": "string",
        },
        "contact_name": {
            "type": "string",
        },
        "contact_email": {
            "type": "string",
        },
        "contact_number": {
            "type": "string",
        },
        "attachments": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "title": {"type": "string"},
                    "url": {"type": "url", "required": True},
                },
            },
        },
    }

    school_directory = {
        "first_name": {
            "type": "string",
        },
        "last_name": {
            "type": "string",
        },
        "email": {
            "type": "string",
        },
        "phone_number": {
            "type": "string",
        },
        "title": {
            "type": "string",
        },
    }
    government_meetings = {
        "title": {
            "type": "string",
        },
        "description": {
            "type": "string",
        },
        "classification": {
            "type": "string",
        },
        "is_cancelled": {
            "type": "boolean",
        },
        "start_time": {
            "type": "datetime",
        },
        "end_time": {
            "type": "datetime",
        },
        "is_all_day_event": {
            "type": "boolean",
        },
        "time_notes": {
            "type": "string",
        },
        "location": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                },
                "address": {
                    "type": "string",
                },
            },
        },
        "links": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "title": {
                        "type": "string",
                    },
                    "url": {
                        "type": "url",
                    },
                },
            },
        },
    }



from typing import Any, Callable, Optional

from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from harambe_core.normalize_url import normalize_url


class ParserTypeUrl:
    def __new__(cls, base_url: Optional[str] = None) -> Any:
        return Annotated[str, AfterValidator(cls.validate_type(base_url))]

    @staticmethod
    def validate_type(base_url: Optional[str]) -> Callable[[str], str]:
        def _validate_type(url: str) -> str:
            return normalize_url(url, base_url)

        return _validate_type



from urllib.parse import urljoin, urlparse, urlunparse


def normalize_url(path: str, base_path: str | None) -> str:
    """
    Normalizes an absolute or relative URL path based on the provided base URL.

    :param base_path: The base URL from which the path was scraped.
    :param path: The absolute or relative path to be normalized.
    :return: Normalized URL.
    """
    path = sanitize_scheme(path)
    validate_allowed_scheme(path)
    if not is_s3_url(path):
        # We append actual URLs at the end of S3 urls occasionally
        # Normalization will turn https:// into http:/
        # TODO: When we handle dynamic downloads in our worker, we need to remove this logic
        #  we should also remove s3 as an allowed scheme all together
        path = _normalize(path)
    escaped_path = path.replace(" ", "%20")

    if base_path is None:
        return escaped_path

    validate_allowed_scheme(base_path, scheme_required=True)

    parsed_base_url = urlparse(base_path, allow_fragments=False)
    return urljoin(parsed_base_url.geturl(), escaped_path)


def _normalize(url: str) -> str:
    parsed_url = urlparse(url)
    return urlunparse(
        (
            parsed_url.scheme,
            parsed_url.netloc.replace("\\", "/"),
            parsed_url.path.replace("\\", "/").replace("//", "/"),
            parsed_url.params,
            parsed_url.query,
            parsed_url.fragment,
        )
    )


def sanitize_scheme(url: str) -> str:
    last_scheme_index = find_highest_index_before_period(url, "/")
    base = "https://"
    if url.startswith("http:"):
        base = "http://"
    return base + url[last_scheme_index + 1 :] if last_scheme_index > 0 else url


s3_scheme = "s3"
allowed_url_schemes = ["http", "https", s3_scheme, "file"]


def is_s3_url(url: str) -> bool:
    return urlparse(url).scheme == s3_scheme


def validate_allowed_scheme(url: str, scheme_required: bool = False) -> None:
    scheme = urlparse(url).scheme
    if (scheme_required and not scheme) or (
        scheme and scheme not in allowed_url_schemes
    ):
        raise ValueError(f"URL scheme {scheme} not allowed.")


def find_highest_index_before_period(s: str, char: str) -> int:
    """
    Find the highest index of a specified character before the first period in a string.

    :param s: The string to search.
    :param char: The character to find.
    :return: The highest index of the character before the first period, or -1
    """
    if not s.startswith("http:") and not s.startswith("https:"):
        return -1

    highest_index = -1
    for i, c in enumerate(s):
        if c == ".":
            break
        if c == char:
            highest_index = i
    return highest_index
